@@ -1,14 +1,13 @@
 //! Backend implementations for integers
 
 use crate::traits::{BitTest, ExactRoots};
-use num_integer::Integer;
 
 #[cfg(feature = "num-bigint")]
 use num_bigint::BigUint;
 #[cfg(feature = "num-bigint")]
-use num_traits::{One, ToPrimitive, Zero};
+use num_traits::ToPrimitive;
 
-macro_rules! impl_bititer_prim {
+macro_rules! impl_bittest_prim {
     ($($T:ty)*) => {$(
         impl BitTest for $T {
             #[inline]
@@ -26,7 +25,7 @@ macro_rules! impl_bititer_prim {
         }
     )*}
 }
-impl_bititer_prim!(u8 u16 u32 u64 u128 usize);
+impl_bittest_prim!(u8 u16 u32 u64 u128 usize);
 
 #[cfg(feature = "num-bigint")]
 impl BitTest for BigUint {
@@ -80,6 +79,7 @@ impl_exactroot_prim!(u8 u16 u32 u64 u128 usize);
 
 #[cfg(feature = "num-bigint")]
 impl ExactRoots for BigUint {
+    // TODO: improve based on SqrtExact.java @ java-math-library
     fn sqrt_exact(&self) -> Option<Self> {
         if (QUAD_RESIDUAL64 >> (self % 64u8).to_u64().unwrap()) & 1 == 0 {
             return None;
