@@ -14,13 +14,17 @@
 use crate::buffer::{NaiveBuffer, PrimeBufferExt};
 use crate::factor::{pollard_rho, squfof};
 use crate::primality::{PrimalityBase, PrimalityRefBase};
-use crate::tables::{MOEBIUS_ODD, SMALL_PRIMES};
+use crate::tables::{MOEBIUS_ODD, SMALL_PRIMES, WHEEL_NEXT, WHEEL_PREV, WHEEL_SIZE};
 #[cfg(feature = "big-table")]
-use crate::tables::{SMALL_PRIMES_INV, SMALL_PRIMES_INVLIM};
+use crate::tables::{SMALL_PRIMES_INV, SMALL_PRIMES_INVLIM, ZETA_LOG_TABLE};
 use crate::traits::{FactorizationConfig, Primality, PrimalityTestConfig, PrimalityUtils};
+use crate::RandPrime;
+#[cfg(feature = "num-bigint")]
+use num_bigint::{BigUint, RandBigInt};
 use num_integer::Roots;
-use num_traits::{FromPrimitive, ToPrimitive};
-use rand::random;
+use num_modular::ModularOps;
+use num_traits::{CheckedAdd, FromPrimitive, Num, RefNum, ToPrimitive};
+use rand::{random, Rng};
 use std::collections::BTreeMap;
 use std::convert::TryFrom;
 
@@ -66,6 +70,15 @@ pub fn is_prime64(target: u64) -> bool {
 /// primality tests are desired, please use is_prime() or PrimeBuffer::is_prime()
 #[cfg(feature = "big-table")]
 pub fn is_prime64(target: u64) -> bool {
+    // shortcuts
+    if target < 1 {
+        return false;
+    }
+    if target & 1 == 0 {
+        return target == 2;
+    }
+
+    // first find in the prime list
     if target < 8167 {
         return SMALL_PRIMES.binary_search(&(target as u16)).is_ok();
     }
@@ -266,14 +279,12 @@ pub fn nprimes(count: usize) -> Vec<u64> {
 }
 
 /// This function re-exports [NaiveBuffer::prime_pi()]
-pub fn prime_pi(limit: u64) -> usize {
-    // TODO (v0.2): Implement stand alone prime_pi with Meissel-Lehmer method
+pub fn prime_pi(limit: u64) -> u64 {
     NaiveBuffer::new().prime_pi(limit)
 }
 
 /// This function re-exports [NaiveBuffer::nth_prime()]
-pub fn nth_prime(n: usize) -> u64 {
-    // TODO (v0.2): Implement stand alone nth_prime with prime_pi and next_prime/prev_prime
+pub fn nth_prime(n: u64) -> u64 {
     NaiveBuffer::new().nth_prime(n)
 }
 
@@ -285,7 +296,7 @@ pub fn primorial<T: PrimalityBase + std::iter::Product>(n: usize) -> T {
         .product()
 }
 
-/// This function calculate the Möbius function of the input integer
+/// This function calculate the Möbius μ(n) function of the input integer
 ///
 /// If the input integer is very hard to factorize, it's better to use
 /// the [factors()] function to control how the factorization is done.
@@ -425,7 +436,7 @@ pub fn prime_pi_bounds<T: ToPrimitive + FromPrimitive>(target: &T) -> (T, T) {
 }
 
 /// Returns the estimated inclusive bounds (low, high) of the n-th prime. If the result
-/// is larger than maximum of T, None will be returned.
+/// is larger than maximum of `T`, [None] will be returned.
 ///
 /// # Reference:
 /// - \[1] Dusart, Pierre. "Estimates of Some Functions Over Primes without R.H."
@@ -513,20 +524,349 @@ pub fn nth_prime_bounds<T: ToPrimitive + FromPrimitive>(target: &T) -> Option<(T
     }
 }
 
+/// Test if the target is a safe prime with Sophie German's definition. It will use the
+/// strict primality test configuration.
+pub fn is_safe_prime<T: PrimalityBase>(target: &T) -> Primality
+where
+    for<'r> &'r T: PrimalityRefBase<T>,
+{
+    let buf = NaiveBuffer::new();
+    let config = Some(PrimalityTestConfig::strict());
+
+    // test (n-1)/2 first since its smaller
+    let sophie_p = buf.is_prime(&(target >> 1), config);
+    if matches!(sophie_p, Primality::No) {
+        return sophie_p;
+    }
+
+    // and then test target itself
+    let target_p = buf.is_prime(target, config);
+    target_p & sophie_p
+}
+
+/// Find the first prime number larger than `target`. If the result causes an overflow,
+/// then [None] will be returned
+#[cfg(not(feature = "big-table"))]
+pub fn next_prime<T: PrimalityBase + CheckedAdd>(
+    target: &T,
+    config: Option<PrimalityTestConfig>,
+) -> Option<T>
+where
+    for<'r> &'r T: PrimalityRefBase<T>,
+{
+    // first search in small primes
+    if let Some(x) = target.to_u8() {
+        let next = match SMALL_PRIMES.binary_search(&target.to_u8().unwrap()) {
+            Ok(pos) => SMALL_PRIMES[pos + 1],
+            Err(pos) => SMALL_PRIMES[pos],
+        };
+        return T::from_u8(next);
+    }
+
+    // then moving along the wheel
+    let mut i = (target % T::from_u8(WHEEL_SIZE).unwrap()).to_u8().unwrap();
+    let mut t = target.clone();
+    loop {
+        let offset = WHEEL_NEXT[i as usize];
+        t = t.checked_add(&T::from_u8(offset).unwrap())?;
+        i = i.addm(offset, &WHEEL_SIZE);
+        if is_prime(&t, config).probably() {
+            break Some(t);
+        }
+    }
+}
+
+/// Find the first prime number larger than `target`. If the result causes an overflow,
+/// then [None] will be returned
+#[cfg(feature = "big-table")]
+pub fn next_prime<T: PrimalityBase + CheckedAdd>(
+    target: &T,
+    config: Option<PrimalityTestConfig>,
+) -> Option<T>
+where
+    for<'r> &'r T: PrimalityRefBase<T>,
+{
+    // first search in small primes
+    if target <= &T::from_u8(255).unwrap() // shortcut for u8
+        || target < &T::from_u16(*SMALL_PRIMES.last().unwrap()).unwrap()
+    {
+        let next = match SMALL_PRIMES.binary_search(&target.to_u16().unwrap()) {
+            Ok(pos) => SMALL_PRIMES[pos + 1],
+            Err(pos) => SMALL_PRIMES[pos],
+        };
+        return T::from_u16(next);
+    }
+
+    // then moving along the wheel
+    let mut i = (target % T::from_u16(WHEEL_SIZE).unwrap())
+        .to_u16()
+        .unwrap();
+    let mut t = target.clone();
+    loop {
+        let offset = WHEEL_NEXT[i as usize];
+        t = t.checked_add(&T::from_u8(offset).unwrap())?;
+        i = i.addm(offset as u16, &WHEEL_SIZE);
+        if is_prime(&t, config).probably() {
+            break Some(t);
+        }
+    }
+}
+
+/// Find the first prime number smaller than `target`. If target is less than 3, then [None]
+/// will be returned.
+#[cfg(not(feature = "big-table"))]
+pub fn prev_prime<T: PrimalityBase>(target: &T, config: Option<PrimalityTestConfig>) -> Option<T>
+where
+    for<'r> &'r T: PrimalityRefBase<T>,
+{
+    if target <= &(T::one() + T::one()) {
+        return None;
+    }
+
+    // first search in small primes
+    if let Some(x) = target.to_u8() {
+        let next = match SMALL_PRIMES.binary_search(&x) {
+            Ok(pos) => SMALL_PRIMES[pos - 1],
+            Err(pos) => SMALL_PRIMES[pos - 1],
+        };
+        return Some(T::from_u8(next).unwrap());
+    }
+
+    // then moving along the wheel
+    let mut i = (target % T::from_u8(WHEEL_SIZE).unwrap()).to_u8().unwrap();
+    let mut t = target.clone();
+    loop {
+        let offset = WHEEL_PREV[i as usize];
+        t = t - T::from_u8(offset).unwrap();
+        i = i.subm(offset, &WHEEL_SIZE);
+        if is_prime(&t, config).probably() {
+            break Some(t);
+        }
+    }
+}
+
+/// Find the first prime number smaller than `target`. If target is less than 3, then [None]
+/// will be returned.
+#[cfg(feature = "big-table")]
+pub fn prev_prime<T: PrimalityBase>(target: &T, config: Option<PrimalityTestConfig>) -> Option<T>
+where
+    for<'r> &'r T: PrimalityRefBase<T>,
+{
+    if target <= &(T::one() + T::one()) {
+        return None;
+    }
+
+    // first search in small primes
+    if target <= &T::from_u8(255).unwrap() // shortcut for u8
+        || target < &T::from_u16(*SMALL_PRIMES.last().unwrap()).unwrap()
+    {
+        let next = match SMALL_PRIMES.binary_search(&target.to_u16().unwrap()) {
+            Ok(pos) => SMALL_PRIMES[pos - 1],
+            Err(pos) => SMALL_PRIMES[pos - 1],
+        };
+        return Some(T::from_u16(next).unwrap());
+    }
+
+    // then moving along the wheel
+    let mut i = (target % T::from_u16(WHEEL_SIZE).unwrap())
+        .to_u16()
+        .unwrap();
+    let mut t = target.clone();
+    loop {
+        let offset = WHEEL_PREV[i as usize];
+        t = t - T::from_u8(offset).unwrap();
+        i = i.subm(offset as u16, &WHEEL_SIZE);
+        if is_prime(&t, config).probably() {
+            break Some(t);
+        }
+    }
+}
+
+macro_rules! impl_randprime_prim {
+    ($($T:ty)*) => {$(
+        impl<R: Rng> RandPrime<$T> for R {
+            #[inline]
+            fn gen_prime(&mut self, bit_size: usize, config: Option<PrimalityTestConfig>) -> $T {
+                if bit_size > (<$T>::BITS as usize) {
+                    panic!("The given bit size limit exceeded the capacity of the integer type!")
+                }
+                let t: $T = self.gen();
+                let t = t >> (<$T>::BITS - bit_size as u32);
+                if is_prime64(t as u64) {
+                    t
+                } else {
+                    // deterministic primality test will be used for integers under u64
+                    match next_prime(&t, None) {
+                        Some(p) => p,
+                        None => self.gen_prime(bit_size, config),
+                    }
+                }
+            }
+
+            #[inline]
+            fn gen_safe_prime(&mut self, bit_size: usize) -> $T {
+                // deterministic primality test will be used for integers under u64
+                let p = self.gen_prime(bit_size, None);
+
+                // test (p-1)/2
+                if is_prime64((p >> 1) as u64) {
+                    return p;
+                }
+                // test 2p+1
+                if let Some(p2) = p.checked_mul(2).and_then(|v| v.checked_add(1)) {
+                    if is_prime64(p2 as u64) {
+                        return p2;
+                    }
+                }
+
+                // try again if failed
+                self.gen_safe_prime(bit_size)
+            }
+        }
+    )*}
+}
+impl_randprime_prim!(u8 u16 u32 u64);
+
+impl<R: Rng> RandPrime<u128> for R {
+    #[inline]
+    fn gen_prime(&mut self, bit_size: usize, config: Option<PrimalityTestConfig>) -> u128 {
+        if bit_size > (u128::BITS as usize) {
+            panic!("The given bit size limit exceeded the capacity of the integer type!")
+        }
+        let t: u128 = self.gen();
+        let t = t >> (u128::BITS - bit_size as u32);
+        if is_prime(&t, config).probably() {
+            t
+        } else {
+            match next_prime(&t, config) {
+                Some(p) => p,
+                None => self.gen_prime(bit_size, config),
+            }
+        }
+    }
+
+    #[inline]
+    fn gen_safe_prime(&mut self, bit_size: usize) -> u128 {
+        let p = self.gen_prime(bit_size, None);
+        let config = Some(PrimalityTestConfig::strict());
+        if is_prime(&(p >> 1), config).probably() {
+            return p;
+        }
+        if let Some(p2) = p.checked_mul(2).and_then(|v| v.checked_add(1)) {
+            if is_prime(&p2, config).probably() {
+                return p2;
+            }
+        }
+
+        self.gen_safe_prime(bit_size)
+    }
+}
+
+#[cfg(feature = "big-int")]
+impl<R: Rng> RandPrime<BigUint> for R {
+    #[inline]
+    fn gen_prime(&mut self, bit_size: usize, config: Option<PrimalityTestConfig>) -> BigUint {
+        let t = self.gen_biguint(bit_size as u64);
+        if is_prime(&t, config).probably() {
+            t
+        } else {
+            match next_prime(&t, config) {
+                Some(p) => p,
+                None => self.gen_prime(bit_size, config),
+            }
+        }
+    }
+
+    #[inline]
+    fn gen_safe_prime(&mut self, bit_size: usize) -> BigUint {
+        let p = self.gen_prime(bit_size, None);
+        let config = Some(PrimalityTestConfig::strict());
+        if is_prime(&(&p >> 1u8), config).probably() {
+            return p;
+        }
+        let p2 = (p << 1u8) + 1u8;
+        if is_prime(&p2, config).probably() {
+            return p2;
+        }
+
+        self.gen_safe_prime(bit_size)
+    }
+}
+
+/// Estimate the value of prime π() function by averaging the estimated bounds.
+#[cfg(not(feature = "big-table"))]
+pub fn prime_pi_est<T: Num + ToPrimitive + FromPrimitive>(target: &T) -> T {
+    let (lo, hi) = prime_pi_bounds(target);
+    (lo + hi) / T::from_u8(2).unwrap()
+}
+
+/// Estimate the value of prime π() function by Riemann's R function. The estimation
+/// error is roughly of scale O(sqrt(x)log(x)).
+///
+/// Reference: <https://primes.utm.edu/howmany.html#better>
+#[cfg(feature = "big-table")]
+pub fn prime_pi_est<T: ToPrimitive + FromPrimitive>(target: &T) -> T {
+    // shortcut
+    if let Some(x) = target.to_u16() {
+        if x <= (*SMALL_PRIMES.last().unwrap()) as u16 {
+            let (lo, hi) = prime_pi_bounds(&x);
+            debug_assert_eq!(lo, hi);
+            return T::from_u16(lo).unwrap();
+        }
+    }
+
+    // Gram expansion with logarithm arithmetics
+    let lnln = target.to_f64().unwrap().ln().ln();
+    let mut total = 0f64;
+    let mut lnp = 0f64; // k*ln(ln(x))
+    let mut lnfac = 0f64; // ln(k!)
+
+    for k in 1usize..100 {
+        lnp += lnln;
+        let lnk = (k as f64).ln();
+        lnfac += lnk;
+        let lnzeta = if k > 64 { 0f64 } else { ZETA_LOG_TABLE[k - 1] };
+        let t = lnp - lnk - lnfac - lnzeta;
+        if t < -4. {
+            // stop if the increment is too small
+            break;
+        }
+        total += t.exp();
+    }
+    T::from_f64(total + 1f64).unwrap()
+}
+
+/// Estimate the value of nth prime by bisecting on [prime_pi_est]
+pub fn nth_prime_est<T: ToPrimitive + FromPrimitive + Num + PartialOrd>(target: &T) -> Option<T>
+where
+    for<'r> &'r T: RefNum<T>,
+{
+    let (mut lo, mut hi) = nth_prime_bounds(target)?;
+    if lo == hi {
+        return Some(lo);
+    }
+
+    while lo != &hi - T::from_u8(1).unwrap() {
+        let x = (&lo + &hi) / T::from_u8(2).unwrap();
+        let mid = prime_pi_est(&x);
+        if &mid < target {
+            lo = x
+        } else if &mid > target {
+            hi = x
+        } else {
+            return Some(x);
+        }
+    }
+    return Some(lo);
+}
+
 // TODO: More functions
 // REF: http://www.numbertheory.org/gnubc/bc_programs.html
 // REF: https://github.com/TilmanNeumann/java-math-library
-// - is_safe_prime: is safe prime with Sophie Germain definition
 // - is_smooth: checks if the smoothness bound is at least b
 // - euler_phi: Euler's totient function
 // - jordan_tot: Jordan's totient function
-// - moebius_mu: Möbius mu function
 // Others include Louiville function, Mangoldt function, Dedekind psi function, Dickman rho function, etc..
-//
-// TODO (v0.2): Implement these prime indexing functions
-// - next_prime (ref `n_primes_next` in FLINT)
-// - prev_prime (similar strategy with `n_primes_next`)
-// - rand_prime
 
 #[cfg(test)]
 mod tests {
@@ -602,10 +942,32 @@ mod tests {
         }
     }
 
+    #[test]
+    fn is_safe_prime_test() {
+        // OEIS:A005385
+        let safe_primes = [
+            5u16, 7, 11, 23, 47, 59, 83, 107, 167, 179, 227, 263, 347, 359, 383, 467, 479, 503,
+            563, 587, 719, 839, 863, 887, 983, 1019, 1187, 1283, 1307, 1319, 1367, 1439, 1487,
+            1523, 1619, 1823, 1907,
+        ];
+        for p in SMALL_PRIMES {
+            let p = p as u16;
+            if p > 1500 {
+                break;
+            }
+            assert_eq!(
+                is_safe_prime(&p).probably(),
+                safe_primes.iter().find(|&v| &p == v).is_some()
+            );
+        }
+    }
+
     #[test]
     fn moebius_mu_test() {
         // test small examples
-        let mu20: [i8; 20] = [1, -1, -1, 0, -1, 1, -1, 0, 0, 1, -1, 0, -1, 1, 1, 0, -1, 0, -1, 0];
+        let mu20: [i8; 20] = [
+            1, -1, -1, 0, -1, 1, -1, 0, 0, 1, -1, 0, -1, 1, 1, 0, -1, 0, -1, 0,
+        ];
         for i in 0..20 {
             assert_eq!(moebius_mu(&(i + 1)), mu20[i], "moebius on {}", i);
         }
@@ -618,14 +980,14 @@ mod tests {
         let sphenic3: [u8; 20] = [
             30, 42, 66, 70, 78, 102, 105, 110, 114, 130, 138, 154, 165, 170, 174, 182, 186, 190,
             195, 222,
-        ]; // OEIS A007304
+        ]; // OEIS:A007304
         for i in 0..20 {
             assert_eq!(moebius_mu(&sphenic3[i]), -1i8, "moebius on {}", sphenic3[i]);
         }
         let sphenic5: [u16; 23] = [
             2310, 2730, 3570, 3990, 4290, 4830, 5610, 6006, 6090, 6270, 6510, 6630, 7410, 7590,
             7770, 7854, 8610, 8778, 8970, 9030, 9282, 9570, 9690,
-        ]; // OEIS A046387
+        ]; // OEIS:A046387
         for i in 0..20 {
             assert_eq!(moebius_mu(&sphenic5[i]), -1i8, "moebius on {}", sphenic5[i]);
         }
@@ -635,6 +997,7 @@ mod tests {
     fn prime_pi_bounds_test() {
         fn check(n: u64, pi: u64) {
             let (lo, hi) = prime_pi_bounds(&n);
+            let est = prime_pi_est(&n);
             assert!(
                 lo <= pi && pi <= hi,
                 "fail to satisfy {} <= pi({}) = {} <= {}",
@@ -642,21 +1005,21 @@ mod tests {
                 n,
                 pi,
                 hi
-            )
+            );
+            assert!(lo <= est && est <= hi);
         }
 
         // test with sieved primes
         let mut pb = NaiveBuffer::new();
         let mut last = 0;
         for (i, p) in pb.primes(100000).cloned().enumerate() {
-            println!("{}, {}", i, p);
             for j in last..p {
                 check(j, i as u64);
             }
             last = p;
         }
 
-        // test with some known cases with input as 10^n, OEIS A006880
+        // test with some known cases with input as 10^n, OEIS:A006880
         let pow10_values = [
             0,
             4,
@@ -695,6 +1058,8 @@ mod tests {
                 p,
                 hi
             );
+            let est = super::nth_prime_est(&n).unwrap();
+            assert!(lo <= est && est <= hi);
         }
 
         // test with sieved primes
@@ -703,7 +1068,7 @@ mod tests {
             check(i as u64 + 1, p as u64);
         }
 
-        // test with some known cases with input as 10^n, OEIS A006988
+        // test with some known cases with input as 10^n, OEIS:A006988
         let pow10_values = [
             2,
             29,
@@ -727,4 +1092,82 @@ mod tests {
             check(n, *nth_prime);
         }
     }
+
+    #[test]
+    fn prev_next_test() {
+        assert_eq!(prev_prime(&2u32, None), None);
+
+        // OEIS:A077800
+        let twine_primes: [(u32, u32); 8] = [
+            (2, 3), // not exactly twine
+            (3, 5),
+            (5, 7),
+            (11, 13),
+            (17, 19),
+            (29, 31),
+            (41, 43),
+            (617, 619),
+        ];
+        for (p1, p2) in twine_primes {
+            assert_eq!(prev_prime(&p2, None).unwrap(), p1);
+            assert_eq!(next_prime(&p1, None).unwrap(), p2);
+        }
+
+        let adj10_primes: [(u32, u32); 7] = [
+            (7, 11),
+            (97, 101),
+            (997, 1009),
+            (9973, 10007),
+            (99991, 100003),
+            (999983, 1000003),
+            (9999991, 10000019),
+        ];
+        for (i, (p1, p2)) in adj10_primes.iter().enumerate() {
+            assert_eq!(prev_prime(p2, None).unwrap(), *p1);
+            assert_eq!(next_prime(p1, None).unwrap(), *p2);
+
+            let pow = 10u32.pow((i + 1) as u32);
+            assert_eq!(prev_prime(&pow, None).unwrap(), *p1);
+            assert_eq!(next_prime(&pow, None).unwrap(), *p2);
+        }
+    }
+
+    #[test]
+    fn rand_prime_test() {
+        let mut rng = rand::thread_rng();
+
+        // test random prime generation for each size
+        let p: u8 = rng.gen_prime(8, None);
+        assert!(is_prime64(p as u64));
+        let p: u16 = rng.gen_prime(16, None);
+        assert!(is_prime64(p as u64));
+        let p: u32 = rng.gen_prime(32, None);
+        assert!(is_prime64(p as u64));
+        let p: u64 = rng.gen_prime(64, None);
+        assert!(is_prime64(p));
+        let p: u128 = rng.gen_prime(128, None);
+        assert!(is_prime(&p, None).probably());
+
+        // test random safe prime generation
+        let p: u8 = rng.gen_safe_prime(8);
+        assert!(is_safe_prime(&p).probably());
+        let p: u32 = rng.gen_safe_prime(32);
+        assert!(is_safe_prime(&p).probably());
+        let p: u128 = rng.gen_safe_prime(128);
+        assert!(is_safe_prime(&p).probably());
+
+        #[cfg(feature = "big-int")]
+        {
+            let p: BigUint = rng.gen_prime(512, None);
+            assert!(is_prime(&p, None).probably());
+            let p: BigUint = rng.gen_safe_prime(192);
+            assert!(is_prime(&p, None).probably());
+        }
+
+        // test bit size limit
+        let p: u16 = rng.gen_prime(12, None);
+        assert!(p < (1 << 12));
+        let p: u32 = rng.gen_prime(24, None);
+        assert!(p < (1 << 24));
+    }
 }
