@@ -1,4 +1,5 @@
 use crate::traits::{BitTest, ExactRoots, PrimalityUtils};
+use either::Either;
 use num_integer::{Integer, Roots};
 use num_modular::ModularOps;
 use num_traits::{FromPrimitive, NumRef, RefNum, ToPrimitive};
@@ -133,10 +134,13 @@ where
         base.powm(&tm1, self).is_one()
     }
 
-    // TODO (v0.2): make is_sprp return Either<bool, T>, to let the test return a composite if possible
-    fn is_sprp(&self, base: T) -> bool {
+    fn is_sprp(&self, base: Self) -> bool {
+        self.test_sprp(base).either(|v| v, |_| false)
+    }
+
+    fn test_sprp(&self, base: T) -> Either<bool, T> {
         if self < &Self::one() {
-            return false;
+            return Either::Left(false);
         }
 
         // find 2^shift*u + 1 = n
@@ -146,17 +150,21 @@ where
 
         let mut x = base.powm(&u, self);
         if x == T::one() || x == tm1 {
-            return true;
+            return Either::Left(true);
         }
 
         for _ in 0..shift {
-            x = (&x).mulm(&x, self);
-            if x == tm1 {
-                return true;
+            let y = (&x).mulm(&x, self);
+            if y.is_one() {
+                return Either::Right(self.gcd(&(x - T::one())));
             }
+            if y == tm1 {
+                return Either::Left(true);
+            }
+            x = y;
         }
 
-        x == T::one()
+        Either::Left(x == T::one())
     }
 
     fn is_lprp(&self, p: Option<usize>, q: Option<isize>) -> bool {
@@ -367,16 +375,19 @@ mod tests {
 
     #[test]
     fn sprp_test() {
-        // strong pseudoprimes of base 2 (OEIS A001262) under 10000
+        // strong pseudoprimes of base 2 (OEIS:A001262) under 10000
         let spsp: [u16; 5] = [2047, 3277, 4033, 4681, 8321];
         for psp in spsp {
             assert!(psp.is_sprp(2));
         }
+
+        // test cofactor return
+        assert!(matches!(341u16.test_sprp(2), Either::Right(31)));
     }
 
     #[test]
     fn lucas_mod_test() {
-        // OEIS A006190
+        // OEIS:A006190
         let p3qm1: [u64; 26] = [
             0,
             1,
@@ -478,7 +489,7 @@ mod tests {
             }
         }
 
-        // lucas pseudoprimes (OEIS A217120) under 10000
+        // lucas pseudoprimes (OEIS:A217120) under 10000
         let lpsp: [u16; 9] = [323, 377, 1159, 1829, 3827, 5459, 5777, 9071, 9179];
         for psp in lpsp {
             assert!(
@@ -498,7 +509,7 @@ mod tests {
             assert!(!n.is_lprp(None, None), "lucas prp test on {}", n);
         }
 
-        // strong lucas pseudoprimes (OEIS A217255) under 10000
+        // strong lucas pseudoprimes (OEIS:A217255) under 10000
         let slpsp: [u16; 2] = [5459, 5777];
         for psp in slpsp {
             assert!(
@@ -518,7 +529,7 @@ mod tests {
             assert!(!n.is_slprp(None, None), "strong lucas prp test on {}", n);
         }
 
-        // extra strong lucas pseudoprimes (OEIS A217719) under 10000
+        // extra strong lucas pseudoprimes (OEIS:A217719) under 10000
         let eslpsp: [u16; 3] = [989, 3239, 5777];
         for psp in eslpsp {
             assert!(
